/* 
 * plcwrite2/plcread2 routines
 *  yow plc (fa-m3) library
 *   k.furukawa, oct.6.1997 - aug.10.2003.
 */
/*
 * this was extracted from plc4.c v3.2
 * please use plc4.c instead of this
 */

static char what_wrdwwr[] =
"@(#)wrdwwr v3.2 yew plc, k.furukawa, oct.1997-aug.2003";

/*
 * plcwrite2 -- write values into a plc (WWR)
 */

int plcwrite2
(
 char *host,	/* host   - host name or ip address of yew plc */
 char *caddr,	/* caddr  - address to write ("Xxxxx") */
 int addr,	/* addr   - address to write (Dxxxx), when caddr is NULL */
 int size,	/* size   - number of words to write (maximum of 256) */
 unsigned short *data	/* data   - data to write */
 )
/* externals: sd, resp, hostsd, s2opt, debugon */
{
    static char template[] = {			/* command to plc */
	0x12,	/* 0: write words sequentially		*/
	0x01,	/* 1: CPU No.				*/
	0x00,	/* 2: sentence consisting of (8 + 2n) bytes	*/
	0x00,	/* 3: 8+n*2  ( cmnd[4] -> cmnd[-] )	*/
	0x00, 0x04,		/* 4: kind of device = D (0x0004)	*/
	0, 0, 0x03, 0xe8,	/* 6: start at channel No.00001000	*/
	0x00, 0x09,		/* 10: n				*/
	0, 0,			/* 12: value 16bit			*/
	0, 0,			/* 14: value 16bit			*/
	0x0
    };
    static unsigned short cmnd[MAX_MSG/2];	/* command for plc */
    int slen = size*2+12;		/* command length */
    int rlen;				/* reply length */

    if ((size>MAX_MSG/4)||(size<1))  goto E_SIZE;

    if ( cmnd[0] != htons(0x1201) )  {		/* first time */
	cmnd[0] = htons(0x1201);		/* command & cpu number */
	cmnd[3] = htons(0x0000);		/* data address uppper word */
    }
    if ( sd == -1 || strncmp(host,hostsd,64) )  {
	s2_close(sd,S2_OERR);
	if((sd = s2_open(YEW_UDP_PORT,"","udp",host,s2opt )) < 0)  {
	    goto E_OPEN;			/* open communication */
	}
	strncpy(hostsd,host,64);		/* save hostname */
    }
    
    cmnd[1] = htons(8+size*2);			/* message size */
    cmnd[5] = htons(size);			/* data number */
    
    if (caddr != NULL)  {			/* address in "Xxxxx" format */
	char c;
	int d;
	if ( 2 != sscanf(caddr, "%c%d", &c, &d) )  
	    goto E_ADDR;			/* address format */
	cmnd[2] = htons(c-'@');			/* data class */
	cmnd[4] = htons(d);			/* data address */
    }  else  {					/* Dxxxx address */
	cmnd[2] = htons(0x0004);		/* data class */
	cmnd[4] = htons(addr);			/* data address */
    }
    
    {
	int i;
	for (i=0; i<size; i++)  {
	    cmnd[i+6] = htons(data[i]);
	}
    }

    if ( (rlen = s2_tsend(sd, (char *)cmnd, slen, RESP_TIME_MS, s2opt))
	!= slen){
	s2_close(sd,S2_OERR);			/* try again */
	if((sd = s2_open(YEW_UDP_PORT,"","udp",host,s2opt)) < 0)  {
	    goto E_OPEN;			/* open communication */
	}
	if ((rlen = s2_tsend(sd,(char *)cmnd,slen,RESP_TIME_MS,s2opt))
	    != slen )  {
	    if (rlen == S2_TIMEOUT)  {		/* time out ? */
		goto E_TIMEOUT;			/* time out error */
	    }  else  {
		goto E_NETWORK;			/* other network error */
	    }
	}
    }
    if ( (rlen = s2_treceive(sd,(char *)resp,MAX_MSG,RESP_TIME_MS,s2opt))
	< 0)  {
	if (rlen == S2_TIMEOUT)
	    goto E_TIMEOUT;			/* time out error */
	else
	    goto E_NETWORK;			/* other network error */
    }
    /* s2_close(sd,s2opt); */
    
    if(ntohs(resp[0]) != 0x9200){		/* failure reponse */
	goto E_NETWORK2;			/* other network error */
    }
    return 0;

 E_SIZE:
    if(debugon)
	printf("plcwrite -- illegal data size '%d'\n", size);
    return -1;
    
 E_ADDR:
    if(debugon)
	printf("plcwrite -- illegal address specification '%s'\n", caddr);
    return -2;
    
 E_OPEN:
    if(debugon)
	printf("plcwrite -- open failure on '%s'\n", host);
    s2_close( sd, s2opt );
    sd = -1;
    return -3;
    
 E_TIMEOUT:
    if(debugon)
	printf("plcwrite -- timeout at send on '%s'\n", host);
    s2_close( sd, s2opt );
    sd = -1;
    return -4;

 E_NETWORK:
    if(debugon)
	printf("plcwrite -- plc network failure on '%s'\n", host);
    s2_close( sd, s2opt );
    sd = -1;
    return -5;

 E_NETWORK2:
    if(debugon)  {
	printf("plcwrite -- plc network internal failure on '%s'\n", host);
	printf(" -- terminal code is %x\n", ntohs(resp[0])&0xff);
    }
    s2_close( sd, s2opt );
    sd = -1;
    return -6;
}


/*
 * plcread2 -- read values from a plc (WRD)
 */
int plcread2
(
 char *host,	/* host   - host name or ip address of yew plc */
 char *caddr,	/* caddr  - address to read ("Yxxxx") */
 int addr,	/* addr   - address to read (Dxxxx), when caddr is NULL */
 int size,	/* size   - number of words to read (maximum of 247?) */
 unsigned short *data	/* data   - data to read */
 )
/* externals: sd, resp, hostsd, s2opt, debugon */
{
    static char template[] = {			/* command to plc */
	0x11,	/* 0: read words sequentially	*/
	0x01,	/* 1: CPU No.					*/
	0x00,	/* 2: sentence consisting of 18 bytes*/
	0x08,	/* 3: 8 ( cmnd[4] -> cmnd[-] )	*/
	0x00, 0x04,		/* 4: kind of device = D (0x0004)	*/
	0, 0, 0x03, 0xe8,	/* 6: start at channel, ex. No.00005000 */
	0x00, 0x00		/* 10: n */
    };
    static unsigned short cmnd[MAX_MSG/2]; /* command for plc */
    int slen = 12;				/* command length */
    int rlen;					/* reply length */

    if ((size>MAX_MSG/4)||(size<1))  goto E_SIZE;

    if ( cmnd[0] != htons(0x1101) )  {		/* first time */
	cmnd[0] = htons(0x1101);		/* command & cpu number */
	cmnd[1] = htons(0x0008);		/* message size */
	cmnd[3] = htons(0x0000);		/* data address upper word */
    }
    if ( sd == -1 || strncmp(host,hostsd,64) )  {
	s2_close(sd,S2_OERR);
	if((sd = s2_open(YEW_UDP_PORT,"","udp",host,s2opt )) < 0)  {
	    goto E_OPEN;
	}
	strncpy(hostsd,host,64);		/* save hostname */
    }

    cmnd[5] = htons(size);			/* data number */

    if (caddr != NULL)  {			/* address in "Xxxxx" format */
	char c;
	int d;
	if ( 2 != sscanf(caddr, "%c%d", &c, &d) )
	    goto E_ADDR;			/* address format */
	cmnd[2] = htons(c-'@');			/* data class */
	cmnd[4] = htons(d);			/* data address */
    }  else  {
	cmnd[2] = htons(0x0004);		/* data class */
	cmnd[4] = htons(addr);			/* data address */
    }

    if (debugon >= 2) plcdumpdata(cmnd,slen,"messge to be sent");
    
    if ((rlen = s2_tsend(sd,(char *)cmnd,slen,RESP_TIME_MS,s2opt))
	!= slen)  {
	s2_close(sd,S2_OERR);			/* try again */
	if((sd = s2_open(YEW_UDP_PORT,"","udp",host,s2opt )) < 0)  {
	    goto E_OPEN;			/* open communication */
	}
	if ((rlen = s2_tsend(sd,(char *)cmnd,slen,RESP_TIME_MS,s2opt))
	    != slen)  {
	    if (rlen == S2_TIMEOUT)  {		/* time out ? */
		goto E_TIMEOUT;			/* time out error */
	    }  else  {
		goto E_NETWORK;			/* other network error */
	    }
	}
    }
    if ((rlen = s2_treceive(sd,(char *)resp,MAX_MSG,RESP_TIME_MS,s2opt))
	< 0)  {
	if (rlen == S2_TIMEOUT)
	    goto E_TIMEOUT;			/* time out error */
	else
	    goto E_NETWORK;			/* other network error */
    }
    /* s2_close( sd, s2opt ); */
    
    if (debugon >= 2) plcdumpdata(resp,rlen,"messge received");
    
    if(ntohs(resp[0]) != 0x9100)  {		/* failure reponse */
	goto E_NETWORK2;			/* other network error */
    }

    {
	int i;
	for (i=0; i<size; i++)  {
	    data[i] = ntohs(resp[i+2]);
	}
    }

    return 0;

 E_SIZE:
    if(debugon)
	printf("plcread -- illegal data size '%d'\n", size);
    return -1;
    
 E_ADDR:
    if(debugon)
	printf("plcwrite -- illegal address specification '%s'\n", caddr);
    return -2;
    
 E_OPEN:
    if(debugon)
	printf("plcread -- open failure on '%s'\n", host);
    s2_close( sd, s2opt );
    sd = -1;
    return -3;
    
 E_TIMEOUT:
    if(debugon)
	printf("plcread -- timeout at send on '%s'\n", host);
    s2_close( sd, s2opt );
    sd = -1;
    return -4;

 E_NETWORK:
    if(debugon)
	printf("plcread -- plc network failure on '%s'\n", host);
    s2_close( sd, s2opt );
    sd = -1;
    return -5;

 E_NETWORK2:
    if(debugon)  {
	printf("plcread -- plc network internal failure on '%s'\n", host);
	printf(" -- terminal code is %x\n", ntohs(resp[0])&0xff);
    }
    s2_close( sd, s2opt );
    return -6;
}
